/*

|--------------------------------------------------------------------------|
|	RBFTRV.C V1.0.d: Radial Basis Function Neural Net Training/Validation  |
|   program. Training algorithm: K-means and/or Stochastic Gradient        |
|                                                                          |
|            By Cristina & Fernando de Castro - October  2001              |
|	                                                                       |
|	V1.0 (October 2001) for DJGPP compiler                                 |
|  	V1.0.d (February 2013) adapted for DevC++ compiler with conio2 library |
|--------------------------------------------------------------------------|

		|---------------------------------------|
		|                                       |
		|   This program should be compiled     |
		|    with DevC++ compiler V 5.3.0.4     | 
		|      using conio2 library             |
		|                                       |
        |---------------------------------------|
*/

/***************************************************************************
*
* HEADERS:
*
****************************************************************************/
#include<conio2.h>
#include<stdio.h>
#include<stdlib.h>
#include<process.h>
#include<math.h>
#include<float.h>
#include<ctype.h>
#include<string.h>
#include<time.h>
#include<dos.h>
#include<fcntl.h>
#include<share.h>


/*****************************************************************************
*
* PROGRAM DEFINITIONS:
*
*****************************************************************************/
#define BUFSIZE 0xFFFF /* file I/O buffer size: 0xFFFF = 64Kb */
#define SMALLBUFSIZE  0x03FF /* file I/O small buffer size 0x03FF = 1Kb  */
#define MAXLINESIZE 65536 /* max line size for  GetSheetNumLines() and NumberOfFloatsInLine() */
#define MAXARGLEN 256   /* max length of an alphanumeric arg */

/*****************************************************************************
*
* COMPILING DIRECTIVES:
*
*****************************************************************************/

//#define DEBUG_FILE "dbg.txt" /* if defined open ascii debug file */

#pragma GCC diagnostic ignored "-Wwrite-strings"

/*****************************************************************************
*
* CONSTANTS:
*
*****************************************************************************/
#define BIAS -1 /* bias level */
#define MAXFLOAT 3.4E38 /* 32 bits DJGPP maximum float */
#define MINABSFLOAT 1E-45 /* 32 bits DJGPP minimum float absolute value regarded as non-zero */

#define MAXEXPARG 88.7 /* 88.7 = ln( 3.4x10^38): max argument value to avoid
			float overflow in function exp() */
#define NALPHALINES 5 /* desired number of alphanumeric lines in parameter file */
#define NFLOATLINES 10 /* desired number of float lines in parameter file */
#define BEL 7  /* ascii BEL */
#define BLINK 128 /* console blink */

/***************************************************************************
*
* DATA TYPE STRUCTURES:
*
****************************************************************************/

typedef struct HiddenNeuron {

float phy;   /*  Hidden Neuron Output */
float NDN;   /*  Normalized-to-variance differential squared Euclidean norm */
float *t;    /*  Hidden Neuron Vector Center (size=Net->Rbf.M) */
float *taux; /*  Hidden Neuron Auxiliary Vector Center (size=Net->Rbf.M) */
float ss;    /*  Hidden Neuron Variance */

}HNEURON;


/*--------------------------------------------------------------------------*/

typedef struct RBFNeuralNetwork {

unsigned N; /* Number of samples vectors in the Training Set TSet -> spec. in TSet input file */
unsigned K; /* Number of Hidden Neurons = Number of RBFs = Number of Gaussian Vector Centers  */
unsigned M; /* Number of Input Nodes = Dimension of each Vector Center -> spec.
               in the TSet file: M=NCols-1 -> Rightmost column of TSet MUST BE
               the Desired Output d column */

float Kn;  /* Vector Centers K-means Learning Rate */
float Wn;  /* Output neuron synapses Stochastic Gradient Learning Rate */
float Tn;  /* Vector Centers coordinates Stochastic Gradient Learning Rate */
float Sn;  /* Vector Centers variance Stochastic Gradient Learning Rate */

float SWIR;   /* Output Neuron Synapses Weight Initialization Range */

HNEURON *Hn; /* Hidden Neurons Storage Area Pointer (size=K) */

float *W;    /* Output Neuron Synapses Vector (size=K) */
float Wb;    /* Output Neuron bias synapse */

float y;    /* RBF Output Neuron output value */
float e;    /* RBF output neuron error e = d - y */

float *AuxM;  /* Temporary Auxiliary Vector (size M) */
float *AuxK;  /* Temporary Auxiliary Vector (size K) */

float KMStopCrit; /* Vector Centers K-means update stopping criterion: The vector centers global percent squared Euclidean norm change w.r.t K-means Learning Rate */

}RBFNN;

/*--------------------------------------------------------------------------*/

typedef struct NeuralNetwork {

RBFNN Rbf; /* Radial Basis Function Neural Network data type */
char *ParameterFile; /* input parameter file */
unsigned NumInputNodes; /* Number of input nodes */
char *TDIF; /* Training Data Input File */
char *VDIF; /* Validation Data input File */
char *FRBFSF; /* Final RBF Specification Output File */
char *MSEOF; /* Training & validation MSE Output File */
FILE *MSEOFp; /* Training & validation MSE Output File Pointer */
char *RBFO2VOF; /* RBF output_to_VDIF output file */
FILE *RBFO2VOFp; /* RBF output_to_VDIF output file */
unsigned TSSize; /* Training Set size (number of vector samples in Training Set */
unsigned VSSize; /* Validation Set size (number of vector samples in Validation Set */
char *buffer; /* file I/O speed up buffer address */
float MSE_T; /* the training MSE after the last epoch */
float MSE_V; /* the validation MSE after the last epoch */
unsigned long NumEpoch; /* the number of epochs until the last epoch */
float MSEChRStopCrit; /* Percent Validation MSE changing rate stopping criterion */
float MSEValStopCrit; /* Validation MSE value stopping criterion */
float **TSet; /* Training Set address */
float **VSet; /* Validation Set address */
float *Avg; /* Training Set average vector */
float *SDev; /* Training Set standard deviation vector */
double A_n,B_n; /* The desired output set D normalization parameters,
                 such that Dn=A_n*D+B_n where Dn lies in the range [-1,1] */
unsigned SmMSEChNEp2C; /* The program assumes CONVERGENCE TO A MINIMUM (hopefully the global)
if the number of nearly consecutive epochs for which the MSE is reducing in a rate  smaller
than MSEChRStopCrit is greater than  SmMSEChNEp2C */
char ValidationMode; /* Flag: if !0 the program executes the training/validation procedure, otherwise it executes only the training procedure */
unsigned K_Means_NumEpoch; /* K-means algorithm number of epochs until convergence */
char Kmeans_On; /* Flag: if !0 K-means algorithm is on scope */

}NEURALNET;

/*--------------------------------------------------------------------------*/

typedef enum {

  NO_EVENT,           /*  " ", */
  MEMORY_ERR,         /*  "Not enough memory", */
  READ_OPEN_ERR,      /*  "I can't find the file", */
  READ_ERR,           /*  "Error reading file", */
  WRITE_OPEN_ERR,     /*  "I can't open the file", */
  WRITE_ERR,          /*  "I can't write the file", */
  BUFFER_ERR,         /*  "I can't attach buffer to file", */
  SIZE_ERR,           /*  "Specified size doesn't match data size in file", */
  MAXIT_ERR,          /*  "Exceeded max number of iterations in function", */
  LOGIC_ERR,          /*  "Logic error in function", */
  READ_NNET_ERR,      /*  "Incompatible neural network in file", */
  MTXDIM_ERR,         /*  "Invalid matrix dimensions in file", */
  PARF_ERR,           /*  "Wrong parameter file", */
  GFTD_ERR,           /*  "I can't get time & date info from file", */
  MAXFLOAT_READ_ERR,  /*  "Exceeded float maximum absolute value reading file" */
  /* ---*/
  USER_INT,	      /*  "Training interrupted by user! Neural Network saved to file", */
  SMALL_MSECHG_INT,   /*  "Training terminated by small enough VMSE Change Rate with VMSE =", */
  SAT_INT,	      /*  "Training terminated due to neurons saturation at VMSE =", */
  SMALL_MSE_INT,      /*  "Training terminated by small enough VMSE =", */
  EARLY_EOF_ERR,      /*  "Found premature EOF in file ", */
  HEADER_SIZE_ERR,    /*  "Found impossible number of samples\nin header of file ", */
  ABORT_INT,          /*  "Training aborted! Neural Network not saved", */
  LNVECSIZE_ERR       /*  "Wrong layer Learning Rate Vector size in file", */

}
EVENT;

/*--------------------------------------------------------------------------*/

typedef struct ParamFileReader{

char *ParameterFile; /* parameter file */
unsigned *FloatLines; /* the array of float lines indexes */
unsigned *AlphaLines; /* the array of alpha lines indexes */
unsigned NumFloatLines; /* the number of float lines */
unsigned NumAlphaLines; /* the number of alpha lines */
char **Sheet; /* parameters file sheet matrix */
char *BufferI; /* input buffer */
FILE *IFp;     /* input filepointer */
unsigned long NumLines;  /* the number of lines read */
unsigned LineMaxLength; /* Sheet max line length */
float **FloatDat; /* float data storage sheet */
char  **AlphaDat;  /* alpha data storage sheet */
unsigned *VectSize; /* the array that contains the size of each row of FloatDat */
char AllocFlag; /* Memory allocation flag: If Allocflag!=0, memory is
                  allocated upon calling Automatic_ParamFileReader() */


}PARFR;


/*****************************************************************************
*
* MACROS:
*
*****************************************************************************/
static double sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

static double aulimarg; /* absolute maximum upper limit */
#define AULIM(a) ( fabs(aulimarg=(a)) >= MAXFLOAT ? ( aulimarg< 0.0 ? (-MAXFLOAT) : (MAXFLOAT)) : aulimarg )


#define SIGNUM(a) ((a) >= 0.0 ? 1.0 : -1.0)


/***************************************************************************
*
* FUNCTION PROTOTYPES:
*
****************************************************************************/
void InitRBF(NEURALNET *Net);
void RowShuffle(float **Mat, unsigned row);
void Help(void);
void ShowParameters(NEURALNET *Net);
void Quit(EVENT Evt, char *Message);
char ChangeParameterMenu(NEURALNET *Net);
void Randomize(void);
float **FloatMatCAlloc(unsigned NRows,unsigned NCols);
void FloatMatFree(float **M);
long GetFileNumSamples(FILE *Fp);
unsigned long GetFileNumLines(FILE *Fp);
float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag);
void FloatMatWrite(float **M,char *file,char *buffer,unsigned NRows,unsigned NCols);
void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines);
void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type);
unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength);
char **CharMatCAlloc(unsigned long NRows,unsigned NCols);
unsigned NumberOfFloatsInLine(char *Line);
float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector);
void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag);
void NetRead(NEURALNET *Net,char *FName);
void NetWrite(NEURALNET *Net, char *FName);
void RBFTrainer(NEURALNET *Net);
void TSetDeAvg (NEURALNET *Net);
void TSetDeSDev(NEURALNET *Net);
void NormalizeDesiredOutput(NEURALNET *Net);
void VSetDeAvg (NEURALNET *Net);
void VSetDeSDev(NEURALNET *Net);
void NormalizeDesiredVSetOutput(NEURALNET *Net);
void QuitPreProcedure(NEURALNET *Net, char *MSE2String);
inline float FloatRandGen(float Range);
inline float GetMinAbsValue(float **VectorSet,unsigned NumVect,unsigned VectDim);
inline float GetMaxAbsValue(float **VectorSet,unsigned NumVect,unsigned VectDim);
inline float DiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim);
inline float SquaredEuclideanNorm(float *Vector,unsigned VectorDim);
void KMeans(NEURALNET *Net);
float MaxSqEuclNormBetweenCenters(NEURALNET *Net);
inline void Phy_O(NEURALNET *Net,float *InputVector);
inline float RBFOutVal(NEURALNET *Net,float *InputVector);
inline void StochGrad(NEURALNET *Net,float *InputVector);
inline void InitRBFVectorCenters(NEURALNET *Net);

/***************************************************************************
*
* GLOBAL VARIABLES:
*
****************************************************************************/

#define NUM_STORED_SCREEN_COORD 5
int XPos[NUM_STORED_SCREEN_COORD]; /* general purpose text screen X coordinates array */
int YPos[NUM_STORED_SCREEN_COORD]; /* general purpose text screen Y coordinates array */


#ifdef DEBUG_FILE
FILE *Dbg; /* debug file */
#endif


/***************************************************************************
*
* MAIN():
*
****************************************************************************/
int main(int argc, char *argv[])
{

NEURALNET Net;

PARFR Pfr;
unsigned Index;
char Type;


unsigned nrows,ncols,ncols_;


/* parameter file lines specification */
unsigned AlphaLines[NALPHALINES]={0,1,2,3,4}; /* see input parameter file */
unsigned FloatLines[NFLOATLINES]={5,6,7,8,9,10,11,12,13,14};/* see input parameter file */


/* Test command line */
if(argc!=2)Help();


#ifdef DEBUG_FILE
if((Dbg=fopen(DEBUG_FILE,"wt"))==NULL)Quit(WRITE_OPEN_ERR,DEBUG_FILE);
#endif

/* store input parameter file name */
Net.ParameterFile=argv[1];


/* Allocate memory to disk I/O speed-up buffer */
Net.buffer=(char *)malloc(BUFSIZE*sizeof(char));
if(!Net.buffer)Quit(MEMORY_ERR,"");


/* Read input parameter file: */

/* Initialize  ParamFileReader */
InitParamFileReader(&Pfr,Net.ParameterFile,FloatLines,AlphaLines,NFLOATLINES,NALPHALINES);


/* Identify the parameter/vector at the i_th line of the parameter file,
 i=0,1,...,NALPHALINES+NFLOATLINES-1 and assign it to the corresponding
 variable. */
ParamStorageID(&Pfr,0,&Index,&Type);
Net.TDIF=Pfr.AlphaDat[Index];/* 0  Training data input file (TDIF) */
ParamStorageID(&Pfr,1,&Index,&Type);
Net.VDIF=Pfr.AlphaDat[Index];/* 1  Validation data input file (VDIF) */
ParamStorageID(&Pfr,2,&Index,&Type);
Net.FRBFSF=Pfr.AlphaDat[Index];/* 2  Final RBF specification output file  */
ParamStorageID(&Pfr,3,&Index,&Type);
Net.MSEOF=Pfr.AlphaDat[Index];/* 3  RBF output_to_TDIF_&_VDIF Mean Square Error (MSE) output file */
ParamStorageID(&Pfr,4,&Index,&Type);
Net.RBFO2VOF=Pfr.AlphaDat[Index]; /* 4  RBF output_to_VDIF output file */
ParamStorageID(&Pfr,5,&Index,&Type);
Net.MSEValStopCrit=Pfr.FloatDat[Index][0];/* 5  Validation MSE (VMSE) value Stopping Criterion (StC) */
ParamStorageID(&Pfr,6,&Index,&Type);
Net.MSEChRStopCrit=Pfr.FloatDat[Index][0];/* 6  Absolute Percent VMSE Changing Rate StC (APMChRStC) */
ParamStorageID(&Pfr,7,&Index,&Type);
Net.SmMSEChNEp2C=Pfr.FloatDat[Index][0];/* 7  The number of consecutive VMSE Changing Rate values below APMChRStC for which the program assumes convergence to a minimum */
ParamStorageID(&Pfr,8,&Index,&Type);
Net.Rbf.SWIR=Pfr.FloatDat[Index][0];/* 8  Output neuron synapses random initialization range */
ParamStorageID(&Pfr,9,&Index,&Type);
Net.Rbf.Kn=Pfr.FloatDat[Index][0];/* 9  Vector Centers coordinates -  K-means update Learning Rate */
ParamStorageID(&Pfr,10,&Index,&Type);
Net.Rbf.KMStopCrit=Pfr.FloatDat[Index][0];/* 10 K-means update stopping criterion: The vector centers global percent squared Euclidean norm change w.r.t K-means Learning Rate */
ParamStorageID(&Pfr,11,&Index,&Type);
Net.Rbf.Tn=Pfr.FloatDat[Index][0];/* 11 Vector Centers coordinates - Stochastic Gradient update Learning Rate */
ParamStorageID(&Pfr,12,&Index,&Type);
Net.Rbf.Sn=Pfr.FloatDat[Index][0];/* 12 Vector Centers variance - Stochastic Gradient update Learning Rate */
ParamStorageID(&Pfr,13,&Index,&Type);
Net.Rbf.Wn=Pfr.FloatDat[Index][0];/* 13 Output neuron synapses - Stochastic Gradient update Learning Rate */
ParamStorageID(&Pfr,14,&Index,&Type);
Net.Rbf.K=Pfr.FloatDat[Index][0]; /* 14 The number of Gaussian Vector Centers */

#if(0)
puts(Net.TDIF);/* 0  Training data input file (TDIF) */
puts(Net.VDIF);/* 1  Validation data input file (VDIF) */
puts(Net.FRBFSF);/* 2  Final RBF specification output file  */
puts(Net.MSEOF);/* 3  RBF output_to_TDIF_&_VDIF Mean Square Error (MSE) output file */
puts(Net.RBFO2VOF);/* 4  RBF output_to_VDIF output file */
printf("%G\n",Net.MSEValStopCrit);/* 5  Validation MSE (VMSE) value Stopping Criterion (StC) */
printf("%G\n",Net.MSEChRStopCrit);/* 6  Absolute Percent VMSE Changing Rate StC (APMChRStC) */
printf("%u\n",Net.SmMSEChNEp2C);/* 7  The number of consecutive VMSE Changing Rate values below APMChRStC for which the program assumes convergence to a minimum */
printf("%G\n",Net.Rbf.SWIR);/* 8  Output neuron synapses random initialization range */
printf("%G\n",Net.Rbf.Kn);/* 9  Vector Centers coordinates -  K-means update Learning Rate */
printf("%G\n",Net.Rbf.KMStopCrit);/* 10 K-means update stopping criterion: The vector centers global percent squared Euclidean norm change w.r.t K-means Learning Rate */
printf("%G\n",Net.Rbf.Tn);/* 11 Vector Centers coordinates - Stochastic Gradient update Learning Rate */
printf("%G\n",Net.Rbf.Sn);/* 12 Vector Centers variance - Stochastic Gradient update Learning Rate */
printf("%G\n",Net.Rbf.Wn);/* 13 Output neuron synapses - Stochastic Gradient update Learning Rate */
printf("%u\n",Net.Rbf.K);/* 14 The number of Gaussian Vector Centers */
exit(0);
#endif


/* Initialize the random number generator random() with a random value */
Randomize();


/* Read Training Set */
Net.TSet=FloatMatReadCheck(Net.TSet=NULL,Net.TDIF,Net.buffer,&nrows,&ncols,1);

/* Architecture parameters */
Net.NumInputNodes=Net.Rbf.M=ncols-1;/* because the last column in Net->TSet is assumed to be the Desired Output */
Net.TSSize=Net.Rbf.N=nrows;

/* Initialize Net */
InitRBF(&Net);

/* Normalize desired output set to range [-1,1] */
NormalizeDesiredOutput(&Net);

/* De-averages Training Set */
TSetDeAvg(&Net);

/* Normalize each TSet input vector component by the corresponding
component of the TSet standard deviation vector. */
TSetDeSDev(&Net);

/* Write normalized/de-averaged/de-deviated input set to disk */
#if(0)
FloatMatWrite(Net.TSet,"tset#.txt",Net.buffer,Net.TSSize,ncols);
#endif


if(Net.ValidationMode){

/* Read Validation Set */
Net.VSet=FloatMatReadCheck(Net.VSet=NULL,Net.VDIF,Net.buffer,&nrows,&ncols_,1);

if(ncols!=ncols_)Quit(SIZE_ERR,Net.VDIF);
Net.VSSize=nrows;

/*  Normalize VSet desired output set to range [-1,1] */
NormalizeDesiredVSetOutput(&Net);

/* De-averages Validation Set */
VSetDeAvg(&Net);

/* Normalize each VSet input vector component by the corresponding
component of the TSet standard deviation vector. */
VSetDeSDev(&Net);

/* Write normalized/de-averaged/de-deviated validation set to disk */
#if(0)
FloatMatWrite(Net.VSet,"vset#.txt",Net.buffer,Net.VSSize,ncols);
#endif
}
else{
Net.VSet=Net.TSet;
Net.VSSize=Net.TSSize;
}


/* Show parameters from parameter file */
Net.Kmeans_On=1;/* Status screen flag: K-means is activated */
ShowParameters(&Net);

/* RBF training loop */
RBFTrainer(&Net);

return 0;
}

/***************************************************************************
* FUNC: void InitRBF(NEURALNET *Net)
*
* DESC: Initialize NEURALNET Net.
****************************************************************************/
void InitRBF(NEURALNET *Net)
{

unsigned i;

/* check validation mode */
if(!strcmp(Net->TDIF,Net->VDIF))Net->ValidationMode=0;
else Net->ValidationMode=1;


/* Alloc TSet average vector */
Net->Avg=(float *)calloc((unsigned)Net->NumInputNodes,sizeof(float));
if(!Net->Avg)Quit(MEMORY_ERR,"");


/* Alloc TSet standard deviation vector */
Net->SDev=(float *)calloc((unsigned)Net->NumInputNodes,sizeof(float));
if(!Net->SDev)Quit(MEMORY_ERR,"");


/* Alloc memory for the RBF output neuron W vector */
Net->Rbf.W = (float *)calloc((unsigned long)(Net->Rbf.K),sizeof(float));
if(!Net->Rbf.W)Quit(MEMORY_ERR,"");


/* Alloc memory for auxiliary vectors */
Net->Rbf.AuxM = (float *)calloc((unsigned long)(Net->Rbf.M),sizeof(float));
if(!Net->Rbf.AuxM)Quit(MEMORY_ERR,"");

Net->Rbf.AuxK = (float *)calloc((unsigned long)(Net->Rbf.K),sizeof(float));
if(!Net->Rbf.AuxK)Quit(MEMORY_ERR,"");


/* Alloc memory for Hidden Neuron Layer storage array  */
Net->Rbf.Hn = (HNEURON *)calloc((unsigned long)(Net->Rbf.K),sizeof(HNEURON));
if(!Net->Rbf.Hn)Quit(MEMORY_ERR,"");

/* Alloc memory for Vector Centers t and taux  */
for(i=0;i< Net->Rbf.K;i++){
Net->Rbf.Hn[i].t = (float *)calloc((unsigned long)(Net->Rbf.M),sizeof(float));
if(!Net->Rbf.Hn[i].t)Quit(MEMORY_ERR,"");

Net->Rbf.Hn[i].taux = (float *)calloc((unsigned long)(Net->Rbf.M),sizeof(float));
if(!Net->Rbf.Hn[i].taux)Quit(MEMORY_ERR,"");

}

/* Randomly init the RBF output neuron synapses transmittance */
for(i=0;i< Net->Rbf.K;i++){
Net->Rbf.W[i]=FloatRandGen(Net->Rbf.SWIR);
}
Net->Rbf.Wb=FloatRandGen(Net->Rbf.SWIR);

}

/***************************************************************************
* FUNC: void RowShuffle(float  **Mat, unsigned NRows)
*
* DESC: Randomly shuffles the rows of matrix Mat.
****************************************************************************/
void RowShuffle(float  **Mat, unsigned NRows)
{

register unsigned i;
unsigned r1;
unsigned r2;

float  *trow;


for(i=0;i<NRows;i++)
{

r1=rand()%NRows;
r2=rand()%NRows;


trow=Mat[r1];
Mat[r1]=Mat[r2];
Mat[r2]=trow;

}

}


/***************************************************************************
* FUNC: void ShowParameters(NEURALNET *Net)
*
* DESC: Show RBF parameters on screen.
****************************************************************************/
void ShowParameters(NEURALNET *Net)
{

system("cls");
int TmpY;
char TmpStr[128];

strcpy(TmpStr,"RBF Neural Network Training/Validation Program V1.0.d");
gotoxy((80-strlen(TmpStr))/2,2);textcolor(YELLOW);cprintf(TmpStr);

gotoxy(1,4);
textcolor(LIGHTCYAN);cprintf(" Training Data Input File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->TDIF));
textcolor(LIGHTCYAN);cprintf(" Validation Data Input File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->VDIF));
textcolor(LIGHTCYAN);cprintf(" Final RBF Specification Output File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->FRBFSF));
textcolor(LIGHTCYAN);cprintf(" Training MSE & Validation MSE (TMSE & VMSE) Output File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->MSEOF));
textcolor(LIGHTCYAN);cprintf(" Validation MSE (VMSE) value Stopping Criterion (StC): ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->MSEValStopCrit);
textcolor(LIGHTCYAN);cprintf(" VMSE Changing Rate StC (VMSEChRStC): ");textcolor(LIGHTGRAY);cprintf("%-8.3G%c\n\r",-1*Net->MSEChRStopCrit,'%');/* -Net->MSEChRStopCrit because it's a diminishing changing rate */
textcolor(LIGHTCYAN);cprintf(" Number of Consecutive VMSEChRStC Attendance Stopping Criterion: ");textcolor(LIGHTGRAY);cprintf("%u\n\r",Net->SmMSEChNEp2C);
textcolor(LIGHTCYAN);cprintf(" Output neuron synapses random initialization range: ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->Rbf.SWIR);
textcolor(LIGHTCYAN);cprintf(" Vector Centers - K-means update Learning Rate (KLR): ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->Rbf.Kn);
textcolor(LIGHTCYAN);cprintf(" K-means StC - centers sq. Euclidean norm chg. w.r.t. KLR: ");textcolor(LIGHTGRAY);cprintf("%-8.3G%c\n\r",Net->Rbf.KMStopCrit,'%');
textcolor(LIGHTCYAN);cprintf(" Vector Centers - Stochastic Gradient (SG) update Learning Rate: ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->Rbf.Tn);
textcolor(LIGHTCYAN);cprintf(" Vector Centers variance - SG update Learning Rate: ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->Rbf.Sn);
textcolor(LIGHTCYAN);cprintf(" Output neuron synapses - SG update Learning Rate: ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->Rbf.Wn);
textcolor(LIGHTCYAN);cprintf(" The number of Gaussian Vector Centers: ");textcolor(LIGHTGRAY);cprintf("%u\n\r",Net->Rbf.K);
textcolor(LIGHTCYAN);cprintf(" Training Set Size / Validation Set Size: ");textcolor(LIGHTGRAY);cprintf("%u/%u\n\r",Net->TSSize,Net->VSSize);

TmpY=wherey()+1;

textcolor(LIGHTMAGENTA);

gotoxy(1,TmpY);if(Net->Kmeans_On){cprintf(" K-Means Epoch:");XPos[0]=wherex();YPos[0]=wherey();}
               else{cprintf(" Stoch-G Epoch:");XPos[0]=wherex();YPos[0]=wherey();}

gotoxy(1,TmpY+1);cprintf(" TMSE:");XPos[2]=wherex();YPos[2]=wherey();

gotoxy(1,TmpY+2);cprintf(" VMSE:");XPos[3]=wherex();YPos[3]=wherey();

gotoxy(32,TmpY+2);cprintf("VMSE Change Rate:");XPos[4]=wherex();YPos[4]=wherey();

gotoxy(32,TmpY+1);cprintf("VMSE Change: ");XPos[1]=wherex();YPos[1]=wherey();


textcolor(WHITE);
}



/***************************************************************************
* FUNC: void Help(void)
*
* DESC: Show the help menu on screen.
****************************************************************************/

void Help(void)
{

char TmpStr[128];

system("cls");
strcpy(TmpStr,"RBF Neural Network Training/Validation Program V1.0.d");
gotoxy((80-strlen(TmpStr))/2,2);textcolor(YELLOW);cprintf(TmpStr);

strcpy(TmpStr,"By Fernando & Cristina de Castro - October 2001");
gotoxy((80-strlen(TmpStr))/2,4);textcolor(LIGHTBLUE);cprintf(TmpStr);

gotoxy(1,7);textcolor(LIGHTCYAN);cputs("Use: Rbftrv ParameterFile\n");
textcolor(LIGHTRED);cputs("\n\rNote: The program assumes the rightmost column in the ascii  training data\n\r        input file matrix is the desired output values column.\n\r");
textcolor(LIGHTGRAY);
exit(1);
}


/***************************************************************************
* FUNC: char ChangeParameterMenu(NEURALNET *Net)
*
* DESC: Change RBF parameters "on the fly". Return the performed change
*       operation.
****************************************************************************/
char ChangeParameterMenu(NEURALNET *Net)
{

#define CGETS_NCH 80 /* max number of chars read by cgets() */

register unsigned i;
int xpos[8],ypos[8];
char TmpStr[128];
char op,op_1;
static char str[CGETS_NCH+3];
char *p;

str[0]=CGETS_NCH+1;  /* There's space for CGETS_NCH chars plus the NULL terminator */
op=op_1=0;
getch();


while(1){

system("cls");

strcpy(TmpStr,"RBF Neural Network Parameters Status:");
gotoxy((80-strlen(TmpStr))/2,1);textcolor(YELLOW);cprintf(TmpStr);
gotoxy(1,2);
textcolor(LIGHTCYAN);cprintf(" RBF Specification File: ");textcolor(LIGHTGRAY);xpos[0]=wherex();ypos[0]=wherey();cprintf("%s\n\r",strupr(Net->FRBFSF));
textcolor(LIGHTCYAN);cprintf(" Validation MSE (VMSE) Value Stopping Criterion: ");textcolor(LIGHTGRAY);xpos[1]=wherex();ypos[1]=wherey();cprintf("%-8.3G\n\r",Net->MSEValStopCrit);
textcolor(LIGHTCYAN);cprintf(" VMSE Changing Rate Stopping Criterion: ");textcolor(LIGHTGRAY);xpos[2]=wherex();ypos[2]=wherey();cprintf("%-8.3G%c\n\r",-1*Net->MSEChRStopCrit,'%');/* -Net->MSEChRStopCrit because it's a diminishing changing rate */
textcolor(LIGHTCYAN);cprintf(" Centers K-Means Learning Rate: ");textcolor(LIGHTGRAY);xpos[3]=wherex();ypos[3]=wherey();cprintf("%-8.3G\n\r",Net->Rbf.Kn);
textcolor(LIGHTCYAN);cprintf(" Centers SG Learning Rate: ");textcolor(LIGHTGRAY);xpos[4]=wherex();ypos[4]=wherey();cprintf("%-8.3G\n\r",Net->Rbf.Tn);
textcolor(LIGHTCYAN);cprintf(" Variance SG Learning Rate: ");textcolor(LIGHTGRAY);xpos[5]=wherex();ypos[5]=wherey();cprintf("%-8.3G\n\r",Net->Rbf.Sn);
textcolor(LIGHTCYAN);cprintf(" Synapses SG Learning Rate: ");textcolor(LIGHTGRAY);xpos[6]=wherex();ypos[6]=wherey();cprintf("%-8.3G\n\r",Net->Rbf.Wn);

strcpy(TmpStr,"Change Menu:");
gotoxy((80-strlen(TmpStr))/2,wherey());textcolor(YELLOW);cprintf(TmpStr);cprintf("%\n\r");
textcolor(LIGHTCYAN);cprintf(" MS");textcolor(LIGHTRED);cprintf("E");textcolor(LIGHTCYAN);cprintf(" Value Stopping Criterion\n\r");
cprintf(" MSE ");textcolor(LIGHTRED);cprintf("C");textcolor(LIGHTCYAN);cprintf("hanging Rate Stopping Criterion\n\r");
if(!Net->Kmeans_On){textcolor(LIGHTRED);cprintf(" R");textcolor(LIGHTCYAN);cprintf("ead Neural Network from file\r\n");}
if(Net->Kmeans_On){textcolor(LIGHTRED);cprintf(" B");textcolor(LIGHTCYAN);cprintf("ypass K-Means and go to Stochastic Gradient\r\n");
textcolor(LIGHTCYAN);cprintf(" Centers ");textcolor(LIGHTRED);cprintf("K");textcolor(LIGHTCYAN);cprintf("-Means Learning Rate\n\r");}
textcolor(LIGHTCYAN);cprintf(" Cen");textcolor(LIGHTRED);cprintf("t");textcolor(LIGHTCYAN);cprintf("ers SG Learning Rate\n\r");
textcolor(LIGHTRED);cprintf(" V");textcolor(LIGHTCYAN);cprintf("ariance SG Learning Rate\n\r");
textcolor(LIGHTRED);cprintf(" S");textcolor(LIGHTCYAN);cprintf("ynapses SG Learning Rate\n\r");
if(!Net->Kmeans_On){textcolor(LIGHTRED);cprintf(" W");textcolor(LIGHTCYAN);cprintf("rite Neural Network to file\r\n");}
textcolor(LIGHTRED);cprintf(" G");textcolor(LIGHTCYAN);cprintf("o ahead\r\n");
if(!Net->Kmeans_On){textcolor(LIGHTRED);cprintf(" Q");textcolor(LIGHTCYAN);cprintf("uit and save Neural Network\r\n");}
textcolor(LIGHTCYAN);cprintf(" Ab");textcolor(LIGHTRED);cprintf("o");textcolor(LIGHTCYAN);cprintf("rt and don't save Neural Network");


gotoxy(1,wherey()+1);textcolor(LIGHTRED);cprintf(" Your Change Menu option?");
cscanf("%c",&op);
op=toupper(op);

textcolor(YELLOW);

switch (op) {
  case 'R':
  textbackground(LIGHTGRAY);
  textcolor(LIGHTRED);
  gotoxy(xpos[0],ypos[0]);
  for(i=0;i<strlen(Net->FRBFSF);i++)cprintf("%c",' ');
  gotoxy(xpos[0],ypos[0]);
  p = cgets(str);
  NetRead(Net,p);
  #if(0)
  NetWrite(Net,"netwrt.txt");
  #endif
  break;
  case 'W':
  textbackground(LIGHTGRAY);
  textcolor(LIGHTRED);
  gotoxy(xpos[0],ypos[0]);
  for(i=0;i<strlen(Net->FRBFSF);i++)cprintf("%c",' ');
  gotoxy(xpos[0],ypos[0]);
  p = cgets(str);
  NetWrite(Net,p);
  break;
  case 'E':
  gotoxy(xpos[1],ypos[1]);
  cscanf("%f",&Net->MSEValStopCrit);
  break;
  case 'C':
  gotoxy(xpos[2],ypos[2]);
  cscanf("%f",&Net->MSEChRStopCrit);
  Net->MSEChRStopCrit=fabs(Net->MSEChRStopCrit);
  break;
  case 'K':
  gotoxy(xpos[3],ypos[3]);
  cscanf("%f",&Net->Rbf.Kn);
  break;
  case 'T':
  gotoxy(xpos[4],ypos[4]);
  cscanf("%f",&Net->Rbf.Tn);
  break;
  case 'V':
  gotoxy(xpos[5],ypos[5]);
  cscanf("%f",&Net->Rbf.Sn);
  break;
  case 'S':
  gotoxy(xpos[6],ypos[6]);
  cscanf("%f",&Net->Rbf.Wn);
  break;
  case 'B':
  return op;
  break;
  case 'G':
  return op_1;
  case 'Q':
  QuitPreProcedure(Net,TmpStr);
  Quit(USER_INT,Net->FRBFSF);
  break;
  case 'O':
  Quit(ABORT_INT,"");
  break;
  default:
  continue;
}

op_1=op;

}/* while(1) */


}



/***************************************************************************
* FUNC: void  Randomize(void)
*
* DESC: Randomize random generator seed.
****************************************************************************/
void Randomize(void)
{
srand((unsigned) time(NULL));
}


/*****************************************************************************
* FUNC: float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
*
* DESC: Allocate memory for a NRows x NCols  float matrix.
*****************************************************************************/
float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
{

float  *x;
float  **Mx;
register unsigned i;


x = (float *)calloc((unsigned long)(NRows*NCols),sizeof(float));
if(!x)Quit(MEMORY_ERR,"");


Mx=(float  **)calloc((unsigned long)NRows,sizeof(float *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(float *)&x[i*NCols];

}

return Mx;
}



/*****************************************************************************
* FUNC: void FloatMatFree(float **M)
*
* DESC: Free memory for the float matrix M.
*****************************************************************************/
void FloatMatFree(float **M)
{

free(M[0]);
free(M);

}


/*****************************************************************************
* FUNC: void Quit(EVENT Evt, char *Message)
*
* DESC: Print Event message and exit to shell.
******************************************************************************/
void Quit(EVENT Evt, char *Message)
{

static char *Event[] = {

/*  NO_EVENT */       " ",
/*  MEMORY_ERR */     "Not enough memory",
/*  READ_OPEN_ERR */  "I can't find the file",
/*  READ_ERR */       "Error reading file",
/*  WRITE_OPEN_ERR */ "I can't open the file",
/*  WRITE_ERR */      "I can't write the file",
/*  BUFFER_ERR */     "I can't attach buffer to file",
/*  SIZE_ERR */       "Specified size doesn't match data size in file",
/*  MAXIT_ERR */      "Exceeded max number of iterations in function",
/*  LOGIC_ERR */      "Logic error in function",
/*  READ_NNET_ERR */  "Incompatible neural network in file",
/*  MTXDIM_ERR */     "Invalid matrix dimensions in file",
/*  PARF_ERR  */      "Wrong parameter file",
/*  GFTD_ERR  */      "I can't get time & date info from file",
/*  MAXFLOAT_READ_ERR */"Exceeded float maximum absolute value reading file",
/*---------*/
/*  USER_INT */	      "Training interrupted by user! Neural Network saved to file",
/*  SMALL_MSECHG_INT */"Training terminated by small enough VMSE Change Rate with VMSE =",
/*  SAT_INT */        "Training terminated due to neurons saturation at VMSE =", 
/*  SMALL_MSE_INT */  "Training terminated by small enough VMSE =",
/*  EARLY_EOF_ERR */  "Found premature EOF in file ",
/*  HEADER_SIZE_ERR */"Found impossible number of samples\nin header of file ",
/*  ABORT_INT */      "Training aborted! Neural Network not saved",
/*  LNVECSIZE_ERR */  "Wrong layer Learning Rate Vector size in file"

};

        
	system("cls");
	textcolor(LIGHTGRAY);

	if (Evt != NO_EVENT) {


		if(Message[0]){
		fprintf (stderr,"%s %s!\n", Event[Evt], Message);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}
		else{
		fprintf (stderr,"%s!\n", Event[Evt]);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}

	}
	exit (0);
}


/*****************************************************************************
* FUNC: unsigned long GetFileNumLines(FILE *Fp)
*
* DESC: Return the number of lines in the text file pointed by Fp.
******************************************************************************/
unsigned long GetFileNumLines(FILE *Fp)
{

  char line[MAXLINESIZE];
  register unsigned long Count;


  Count=0;
  while (fgets(line,MAXLINESIZE-2,Fp)){
  Count++;
  }

  if(strlen(line)==1)Count--;/* to discount the last \n in file
                              for the case the user have typed it */


  rewind(Fp);
  return Count;

}



/*****************************************************************************
* FUNC: long GetFileNumSamples(FILE *Fp)
*
* DESC: Return the number of numeric samples in the text file pointed by Fp.
*       Return -1 on read error.
*****************************************************************************/
long GetFileNumSamples(FILE *Fp)
{

register long Count;
float x;
int test;

Count=0;

while((test=fscanf(Fp,"%f",&x))!=-1){
if(test!=1)return -1;
Count++;
}

rewind(Fp);
return Count;

}


/*****************************************************************************
* FUNC: float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag)
*
* DESC: Read an ascii float matrix from 'file' and store it in matrix
*       M[NRows][NCols]. If AllocFlag==1 the function allocates memory
*       for M and the arguments *NRows and *NCols identify the size of M.
*
*       FloatMatReadCheck() checks the float compatibility of the read data.
*****************************************************************************/
float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag)
{

register unsigned i,j;
FILE *in;
long NumSamples;
unsigned long NumLines;
int Test;
double Value;

if((in=fopen(file,"rt"))==NULL)Quit(READ_OPEN_ERR,file);
if (setvbuf(in, buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,file);

NumSamples=GetFileNumSamples(in);
if(NumSamples==-1)Quit(READ_ERR,file);
NumLines=GetFileNumLines(in);

if(NumSamples%NumLines)Quit(MTXDIM_ERR,file); /* check valid matrix dimensions */

*NRows=NumLines;
*NCols=NumSamples/NumLines;

if(AllocFlag==1)M=FloatMatCAlloc(*NRows,*NCols);

for(i=0;i<*NRows;i++){
for(j=0;j<*NCols;j++){
Test=fscanf(in,"%lf",&Value);
if(Test!=1)Quit(READ_ERR,file);
if(fabs(Value)>MAXFLOAT)Quit(MAXFLOAT_READ_ERR,file);
M[i][j]=Value;
}
}

fclose(in);

return M;
}



/*****************************************************************************
* FUNC: void FloatMatWrite(float **M,char *file,char *buffer,unsigned NRows,unsigned NCols)
*
* DESC: Write a float NRows x NCols matrix to an ascii file. If file=NULL, it
*       writes to stdout.
*****************************************************************************/
void FloatMatWrite(float **M,char *file,char *buffer,unsigned NRows,unsigned NCols)
{

register unsigned i,j;
FILE *out;
int test;

if(file!=NULL){
if((out=fopen(file,"wt"))==NULL)Quit(WRITE_OPEN_ERR,file);
if (setvbuf(out, buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,file);
}
else out=stdout;


for(i=0;i<NRows;i++){
for(j=0;j<NCols;j++){

fprintf(out,"%8.6G\t",M[i][j]);

}

test=fprintf(out,"\n");
if(test!=sizeof(char))Quit(WRITE_ERR,file);

}

if(file!=NULL)fclose(out);


}


/*****************************************************************************
* FUNC: void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines)
*
* DESC: Init parameter file reader.
*****************************************************************************/
void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines)

{

register unsigned i;

Pfr->ParameterFile=ParameterFile;
Pfr->FloatLines=FloatLines;
Pfr->AlphaLines=AlphaLines;
Pfr->NumFloatLines=NumFloatLines;
Pfr->NumAlphaLines=NumAlphaLines;


/* Alloc memory for disk I/O speed-up buffers */
Pfr->BufferI=(char *)calloc(SMALLBUFSIZE,sizeof(char));
if(!Pfr->BufferI)Quit(MEMORY_ERR,"");


/* Open parameter file */
if((Pfr->IFp=fopen(Pfr->ParameterFile,"rt"))==NULL)Quit(READ_OPEN_ERR,Pfr->ParameterFile);
if (setvbuf(Pfr->IFp, Pfr->BufferI, _IOFBF, SMALLBUFSIZE) != 0)Quit(BUFFER_ERR,Pfr->ParameterFile);

/* Get the total number of lines and the max line length */
Pfr->NumLines=GetSheetNumLines(Pfr->IFp,&Pfr->LineMaxLength);

/* check */
if(Pfr->NumLines!=(Pfr->NumFloatLines+Pfr->NumAlphaLines))Quit(PARF_ERR,Pfr->ParameterFile);

/* Alloc memory for Sheet */
Pfr->Sheet=CharMatCAlloc(Pfr->NumLines,Pfr->LineMaxLength);

/* Read sheet and store it into Sheet */
for(i=0;i<Pfr->NumLines;i++)fgets(Pfr->Sheet[i],Pfr->LineMaxLength,Pfr->IFp);

/* close parameter file */
fclose(Pfr->IFp);

/* Alloc memory for float FloatDat row pointers */
Pfr->FloatDat=(float **)calloc(Pfr->NumFloatLines,sizeof(float *));
if(!Pfr->FloatDat)Quit(MEMORY_ERR,"");

/* Alloc memory for alphanumeric data */
Pfr->AlphaDat=CharMatCAlloc(Pfr->NumAlphaLines,MAXARGLEN);

/* Alloc memory for the array that contains the size of each row of FloatDat */
Pfr->VectSize=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->VectSize)Quit(MEMORY_ERR,"");

/* obtain float FloatDat sheet from char Sheet */
for(i=0;i<Pfr->NumFloatLines;i++){
Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],NULL);
if(Pfr->FloatDat[i]==NULL)Quit(PARF_ERR,Pfr->ParameterFile);
}

/* obtain string AlphaDat sheet from first string in Sheet line */
for(i=0;i<Pfr->NumAlphaLines;i++){
sscanf(Pfr->Sheet[Pfr->AlphaLines[i]],"%s",Pfr->AlphaDat[i]);
}


}


/*****************************************************************************
* FUNC: void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type)
*
* DESC: StorageIndex return the index in respective (FloatDat/AlphaDat)
*       storage sheet that corresponds to line LineNumber.
*
*       Type returns the respective data type: 'f'->float 'a'->alpha
*****************************************************************************/
void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type)
{

register unsigned i;
int p=-1;
char ParType=0;

//printf("LineNumber=%u\n",LineNumber);

/* scan FloatLines index array */
for(i=0;i<Pfr->NumFloatLines;i++){
if(Pfr->FloatLines[i]==LineNumber){p=i;ParType='f';break;}
}

/* if not found, scan AlphaLines index array */
if(p==-1){
for(i=0;i<Pfr->NumAlphaLines;i++){
if(Pfr->AlphaLines[i]==LineNumber){p=i;ParType='a';break;}
}

}

/* not found */
if(p==-1){fprintf(stderr,"No such line number in parameter file %s!\n",Pfr->ParameterFile);putch(BEL);exit(1);}

/* found */
*StorageIndex=p;
*Type=ParType;

}


/*****************************************************************************
* FUNC: unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength)
*
* DESC: Return number of lines of the text sheet in file pointed by Fp.
*       MaxLength return the line max length, including the '\0'.
******************************************************************************/
unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength)
{

  char line[MAXLINESIZE];
  register unsigned long Count;
  unsigned Length;

  *MaxLength=0;


  Count=0;
  while (fgets(line,MAXLINESIZE-2,Fp)){
  if((Length=strlen(line))>*MaxLength)*MaxLength=Length;
  Count++;
  }

  if(strlen(line)==1)Count--;/* to discount the last \n in file
                              for the case the user has typed it */

  *MaxLength= (*MaxLength)+1; /* to handle the '\0' added by fgets() */

  rewind(Fp);
  return Count;


}


/*****************************************************************************
* FUNC: char **CharMatCAlloc(unsigned long NRows,unsigned NCols)
*
* DESC: Allocate memory for a NRows x NCols  char matrix.
*****************************************************************************/
char **CharMatCAlloc(unsigned long NRows,unsigned NCols)
{
char  *x;
char  **Mx;
register unsigned  i;

x = (char  *)calloc((unsigned long)(NRows*NCols),sizeof(char));
if(!x)Quit(MEMORY_ERR,"");

Mx=(char  **)calloc((unsigned long)NRows,sizeof(char *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(char *)&x[i*NCols];

}

return Mx;
}


/*****************************************************************************
* FUNC: unsigned NumberOfFloatsInLine(char *Line)
*
* DESC: Return the number of floats in Line, starting from the beginning.
*       Upon finding a non-readable character, the function ends the count.
*****************************************************************************/
unsigned NumberOfFloatsInLine(char *Line)
{
static char Str[MAXLINESIZE];
static char Buf[MAXLINESIZE];

char *NonReadable;
double Val;
unsigned Count=0;

#if(0)
puts(Line);
#endif

while(1){
sprintf(Buf,"%s",Line);
if(Buf[0]=='\t'){Line++;continue;} /* found a tab */
if(Buf[0]==' '){Line++;continue;}  /* found a space */
if(!Buf[0])break; /* found a '\0' */
if(Buf[0]=='\n')break; /* found a '\n' */
sscanf(Buf,"%s",Str);
Val= strtod(Str, &NonReadable);

#if(0)
if(NonReadable[0]&&NonReadable[0]!='i')break; /* Check non-readable except 'i'-> to handle complex numbers */
#else
if(NonReadable[0])break; /* Check non-readable */
#endif

#if(0)
fprintf(stdout,"Val[%u]=%8.6lg%c\n",Count,Val,NonReadable[0]);
#endif

Line+=strlen(Str)+1;
Count++;
}

#if(0)
fprintf(stdout,"Line has %u floats.\n",Count);
#endif

return Count;
}


/*****************************************************************************
* FUNC: float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector)
*
* DESC:  Convert the floats in Line into a float vector Vector, starting from the
*        beginning.
*
*        Upon finding a non-readable character, the function ends the count.
*        VectSize returns the size of Vector. If Vector==NULL, then memory
*        is allocated for the float vector Vector got from Line.
*****************************************************************************/
float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector)
{
static char Str[MAXLINESIZE];
static char Buf[MAXLINESIZE];

char *NonReadable;
double Val;
unsigned Count=0;

*VectSize=NumberOfFloatsInLine(Line);
if(*VectSize==0)return NULL;

if(Vector==NULL){
Vector=(float *)calloc(*VectSize,sizeof(float));
if(!Vector)Quit(MEMORY_ERR,"");
}


while(1){
sprintf(Buf,"%s",Line);
if(Buf[0]=='\t'){Line++;continue;} /* found a tab */
if(Buf[0]==' '){Line++;continue;}  /* found a space */
if(!Buf[0])break; /* found a '\0' */
if(Buf[0]=='\n')break; /* found a '\n' */
sscanf(Buf,"%s",Str);
Val= strtod(Str, &NonReadable);

#if(0)
if(NonReadable[0]&&NonReadable[0]!='i')break; /* Check non-readable except 'i'-> to handle complex numbers */
#else
if(NonReadable[0])break; /* Check non-readable */
#endif

Vector[Count]=AULIM(Val); /* check float absolute upper limit */
if(Vector[Count]!=(float)Val){putch(BEL);fprintf(stderr,"Exceeded float maximum absolute value reading parameter from disk!");exit(1);}
Line+=strlen(Str)+1;
Count++;
}


return Vector;
}

/*****************************************************************************
* FUNC: void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag)
*
* DESC: Reads automatically a parameter file.If AllocFlag!=0, memory is allocated.
*****************************************************************************/
void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag)
{

#define SIZE_GUARD 6 /* a guard size for Pfr->Sheet line length to avoid eventual diferent float read from ParameterFile */

register unsigned i,j,k;

Pfr->AllocFlag=AllocFlag;
Pfr->ParameterFile=ParameterFile;


Pfr->NumFloatLines=0;
Pfr->NumAlphaLines=0;


/* Alloc memory for disk I/O speed-up buffers */
if(AllocFlag){
Pfr->BufferI=(char *)calloc(BUFSIZE,sizeof(char));
if(!Pfr->BufferI)Quit(MEMORY_ERR,"");
}


/* Open parameter file */
if((Pfr->IFp=fopen(Pfr->ParameterFile,"rt"))==NULL)Quit(READ_OPEN_ERR,Pfr->ParameterFile);
if (setvbuf(Pfr->IFp, Pfr->BufferI, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Pfr->ParameterFile);

/* Get the total number of lines and the max line length */
Pfr->NumLines=GetSheetNumLines(Pfr->IFp,&Pfr->LineMaxLength);


/* Alloc memory for Sheet */
if(AllocFlag){
Pfr->Sheet=CharMatCAlloc(Pfr->NumLines,Pfr->LineMaxLength+SIZE_GUARD);
}

/* Read sheet and store it into Sheet */
for(i=0;i<Pfr->NumLines;i++)fgets(Pfr->Sheet[i],Pfr->LineMaxLength,Pfr->IFp);

/* close parameter file */
fclose(Pfr->IFp);


/* get NumFloatLines and NumAlphaLines from char Sheet */
for(i=0;i<Pfr->NumLines;i++){

if(NumberOfFloatsInLine(Pfr->Sheet[i])==0)Pfr->NumAlphaLines++;
else Pfr->NumFloatLines++;

}

/* check */
if(Pfr->NumLines!=(Pfr->NumFloatLines+Pfr->NumAlphaLines))Quit(PARF_ERR,Pfr->ParameterFile);



/* Alloc memory for FloatLines and AlphaLines ID vectors */
if(AllocFlag){
Pfr->FloatLines=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->FloatLines)Quit(MEMORY_ERR,"");

Pfr->AlphaLines=(unsigned *)calloc(Pfr->NumAlphaLines,sizeof(unsigned));
if(!Pfr->AlphaLines)Quit(MEMORY_ERR,"");
}

/* Identify FloatLines and AlphaLines indexes from char Sheet */
j=k=0;
for(i=0;i<Pfr->NumLines;i++){

if(NumberOfFloatsInLine(Pfr->Sheet[i])==0)Pfr->AlphaLines[j++]=i;
else Pfr->FloatLines[k++]=i;

}

/* Alloc memory for float FloatDat row pointers */
if(AllocFlag){
Pfr->FloatDat=(float **)calloc(Pfr->NumFloatLines,sizeof(float *));
if(!Pfr->FloatDat)Quit(MEMORY_ERR,"");

/* Alloc memory for alphanumeric data */
Pfr->AlphaDat=CharMatCAlloc(Pfr->NumAlphaLines,MAXARGLEN);

/* Alloc memory for the array that contains the size of each row of FloatDat */
Pfr->VectSize=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->VectSize)Quit(MEMORY_ERR,"");
}

/* obtain float FloatDat sheet from char Sheet */
for(i=0;i<Pfr->NumFloatLines;i++){

if(AllocFlag)Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],NULL);
else Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],Pfr->FloatDat[i]);

if(Pfr->FloatDat[i]==NULL)Quit(PARF_ERR,Pfr->ParameterFile);
}

/* obtain string AlphaDat sheet from first string in Sheet line */
for(i=0;i<Pfr->NumAlphaLines;i++){
sscanf(Pfr->Sheet[Pfr->AlphaLines[i]],"%s",Pfr->AlphaDat[i]);
}


}



/**************************************************************************
* FUNC: void NetRead(NEURALNET *Net,char *FName)
*
* DESC: Read the RBF specification from an ascii text file FName.
*************************************************************************/
void NetRead(NEURALNET *Net,char *FName)
{

#define INEQUALITY_DET_TRIGGER 0.01

register unsigned i,j,LineCount;

static PARFR PfrS;
static char EntryFlag;
char AllocFlag;
float *FTmpPtr;
unsigned K,M;


/*  AllocFlag=1 on the first call of NetRead(). AllocFlag=0 in
any other instant */
if(!EntryFlag){EntryFlag=AllocFlag=1;}
else {AllocFlag=0;}

/* read neural network from File */
Automatic_ParamFileReader(&PfrS,FName,AllocFlag);


/* Get architecture parameters */
LineCount=0;

K=PfrS.FloatDat[LineCount++][0]; /* LineCount = 0 */

M=PfrS.FloatDat[LineCount++][0]; /* LineCount = 1 */

if((K!=Net->Rbf.K)||M!=Net->Rbf.M)Quit(READ_NNET_ERR,FName);

Net->K_Means_NumEpoch=PfrS.FloatDat[LineCount++][0]; /* LineCount = 2 */

Net->NumEpoch=PfrS.FloatDat[LineCount++][0]; /* LineCount = 3 */

Net->MSE_T=PfrS.FloatDat[LineCount++][0]; /* LineCount = 4 */

Net->MSE_V=PfrS.FloatDat[LineCount++][0]; /* LineCount = 5 */

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 6 */

for(i=0;i<Net->NumInputNodes;i++){
if(fabs(1-Net->Avg[i]/FTmpPtr[i])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);
}

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 7 */

for(i=0;i<Net->NumInputNodes;i++){
if(fabs(1-Net->SDev[i]/FTmpPtr[i])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);
}

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 8 */

if(fabs(1-Net->A_n/FTmpPtr[0])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);
if(fabs(1-Net->B_n/FTmpPtr[1])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);


for(i=0;i<Net->Rbf.K;i++){

FTmpPtr=PfrS.FloatDat[LineCount]; /* LineCount = 9 + i */

for(j=0;j<Net->Rbf.M;j++){
Net->Rbf.Hn[i].taux[j]=FTmpPtr[j];
}/* for j */

LineCount++; /* LineCount increment = i increment = hidden neuron index increment */

}/* for i */


for(i=0;i<Net->Rbf.K;i++){

FTmpPtr=PfrS.FloatDat[LineCount]; /* LineCount = 9 + K + i */

for(j=0;j<Net->Rbf.M;j++){
Net->Rbf.Hn[i].t[j]=FTmpPtr[j];
}/* for j */

LineCount++; /* LineCount increment = i increment = hidden neuron index increment */

}/* for i */

FTmpPtr=PfrS.FloatDat[LineCount++]; /* LineCount = 9 + 2K + i */

for(i=0;i<Net->Rbf.K;i++)Net->Rbf.Hn[i].ss=FTmpPtr[i];

FTmpPtr=PfrS.FloatDat[LineCount++]; /* LineCount = 9 + 2K + i + 1 */

Net->Rbf.Wb=FTmpPtr[0];FTmpPtr++;
for(i=0;i<Net->Rbf.K;i++)Net->Rbf.W[i]=FTmpPtr[i];  /* LineCount = 9 + 2K + i + 2 */


}

/***************************************************************************
* FUNC: void NetWrite(NEURALNET *Net, char *FName)
*
* DESC: Write RBF specification in file FName.
****************************************************************************/
void NetWrite(NEURALNET *Net, char *FName)
{

char ParFileTimeDate[128];

FILE *fp;

register unsigned i,j;


if((fp=fopen(FName,"wt"))==NULL)Quit(WRITE_OPEN_ERR,FName);
if (setvbuf(fp, Net->buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,FName);

fprintf(fp,"%8s\t%c RBFTRV parameter file and its date & time.\n","ParFileTimeDate",'%');
fprintf(fp,"%8u\t%c The number of hidden neurons in the RBF.\n",Net->Rbf.K,'%');
fprintf(fp,"%8u\t%c The number of input nodes in the RBF.\n",Net->NumInputNodes,'%');
fprintf(fp,"%8u\t%c The final number of epochs for the K-Means centers training.\n",Net->K_Means_NumEpoch,'%');
fprintf(fp,"%8lu\t%c The final number of epochs for the Stochastic Gradient training.\n",Net->NumEpoch,'%');
fprintf(fp,"%8.6G\t%c The final training MSE.\n",Net->MSE_T,'%');
fprintf(fp,"%8.6G\t%c The final validation MSE.\n",Net->MSE_V,'%');
fprintf(fp,"The training set average vector:\n");
for(i=0;i<Net->NumInputNodes;i++)fprintf(fp,"%8.6G\t",Net->Avg[i]);
fprintf(fp,"\nThe training set standard deviation vector:\n");
for(i=0;i<Net->NumInputNodes;i++)fprintf(fp,"%8.6G\t",Net->SDev[i]);
fprintf(fp,"\nThe training desired output set D normalization parameters A_n and B_n, such that Dn=A_n*D+B_n where Dn lies in the range [-1,1]:\n");
fprintf(fp,"%8.6G\t%8.6G\t",Net->A_n,Net->B_n);

fprintf(fp,"\nThe %u RBF Centers Vectors determined by K-Means, each one a %u-dimensional vector:\n",Net->Rbf.K,Net->Rbf.M);

for(i=0;i<Net->Rbf.K;i++){
for(j=0;j<Net->Rbf.M;j++){
fprintf(fp,"%8.6G\t",Net->Rbf.Hn[i].taux[j]);
}
fprintf(fp,"\n");
}

fprintf(fp,"The %u RBF Centers Vectors determined by SG, each one a %u-dimensional vector:\n",Net->Rbf.K,Net->Rbf.M);

for(i=0;i<Net->Rbf.K;i++){
for(j=0;j<Net->Rbf.M;j++){
fprintf(fp,"%8.6G\t",Net->Rbf.Hn[i].t[j]);
}
fprintf(fp,"\n");
}
fprintf(fp,"The respective RBF centers Variance:\n");

for(i=0;i<Net->Rbf.K;i++){
fprintf(fp,"%8.6G\t",Net->Rbf.Hn[i].ss);
}
fprintf(fp,"\n");

fprintf(fp,"The RBF output neuron Synapses Vector (leftmost component - Bias Synapse):\n");
fprintf(fp,"%8.6G\t",Net->Rbf.Wb);
for(i=0;i<Net->Rbf.K;i++){
fprintf(fp,"%8.6G\t",Net->Rbf.W[i]);
}

fprintf(fp,"\n");


fclose(fp);

}


/*****************************************************************************
* FUNC: void RBFTrainer(NEURALNET *Net)
*
* DESC: RBF training/validation loop.
*****************************************************************************/
void RBFTrainer(NEURALNET *Net)
{

#define RECTANGLE 22
#define DOWN_TRIANGLE 31
#define UP_TRIANGLE 30

register unsigned i;

float MSE_T_1=0; /* previous epoch Net->MSE_T */
float MSE_V_1=0; /* previous epoch Net->MSE_V */
float MSE_T_ChgRate; /* Training MSE changing rate */
float MSE_V_ChgRate; /* Validation MSE changing rate */
register unsigned long Epoch;  /* epoch counter */
unsigned SatDet; /* neuron saturation detector */
unsigned MSE_V_LockUpDet; /* validaton MSE lock up detector */
unsigned SmallEnoughMSEChgRateCounter; /* small enough MSE change rate counter  */
char MSEStr[128]; /* string to store the MSE value */
float DMax2; /* the squared max Euclidean distance between the vector centers */

/* Open MSE output file */
if((Net->MSEOFp=fopen(Net->MSEOF,"wt"))==NULL)Quit(WRITE_OPEN_ERR,Net->MSEOF);
if (setvbuf(Net->MSEOFp,NULL,_IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Net->MSEOF);

/* Init variables */
Epoch=SatDet=MSE_V_LockUpDet=SmallEnoughMSEChgRateCounter=0;


/* Randomly init RBF Vector Centers */
InitRBFVectorCenters(Net);

/* Find the central coordinates of Net->Rbf.K  clusters in TSet and assign them
 to the RBF vector centers */
KMeans(Net);

/* show parameters in screen */
ShowParameters(Net);

/* Determine the averaged square of the max distance between Centers */
if(Net->Rbf.K>1)DMax2=MaxSqEuclNormBetweenCenters(Net)/Net->Rbf.K;
else DMax2=10*SquaredEuclideanNorm(Net->Rbf.Hn[0].t,Net->Rbf.M);

/* Init Centers Variance (see eq. 20.10 Haykin ) */
for(i=0;i<Net->Rbf.K;i++)Net->Rbf.Hn[i].ss=DMax2;


/* Stochastic Gradient Training/validation loop */
while(1){

/* Initial procedures */
gotoxy(XPos[0],YPos[0]);cprintf("%3lu",Epoch); /* show in Status screen */

Net->NumEpoch=Epoch;

if(Epoch==0){
MSE_V_1=Net->MSE_V;
MSE_T_1=Net->MSE_T;
}

/* Training Set presentation */
Net->MSE_T=0;
for(i=0;i<Net->TSSize;i++)
{
Net->MSE_T+=RBFOutVal(Net, Net->TSet[i]);
StochGrad(Net,Net->TSet[i]);
}

Net->MSE_T/=2*Net->TSSize; /* 1/2 -> J=(1/2)*E{e^2} */

if(MSE_T_1!=0.0)MSE_T_ChgRate=AULIM(Net->MSE_T/MSE_T_1)-1;
else MSE_T_ChgRate=MAXFLOAT*SIGNUM(Net->MSE_T);

/* Validation Set presentation */
if(Net->ValidationMode){
Net->MSE_V=0;
for(i=0;i<Net->VSSize;i++)
{
Net->MSE_V+=RBFOutVal(Net, Net->VSet[i]);
}

Net->MSE_V/=2*Net->VSSize;  /* 1/2 -> J=(1/2)*E{e^2} */
if(MSE_V_1!=0.0)MSE_V_ChgRate=AULIM(Net->MSE_V/MSE_V_1)-1;
else MSE_V_ChgRate=MAXFLOAT*SIGNUM(Net->MSE_V);

}
else{
Net->MSE_V=Net->MSE_T;
MSE_V_ChgRate=MSE_T_ChgRate;
}


/* Write Training & Validation MSE to Net->MSEOFp */
if(Net->ValidationMode)fprintf(Net->MSEOFp,"%-.1E\t%-.1E\n",Net->MSE_T,Net->MSE_V);
else fprintf(Net->MSEOFp,"%-.1E\n",Net->MSE_V);

/* Show in Status screen: */
gotoxy(XPos[2],YPos[2]);cprintf("%-.1E",Net->MSE_T);
gotoxy(XPos[3],YPos[3]);cprintf("%-.1E\r",Net->MSE_V);
gotoxy(XPos[4],YPos[4]);;cprintf(" %-8.2G%c\r",fabs(100*MSE_V_ChgRate),'%');

/* Show validation MSE motion indicator in Status screen */
gotoxy(XPos[1],YPos[1]);
if(MSE_V_ChgRate==0.0){
textcolor(YELLOW);
cprintf("%c",RECTANGLE);
}
else if(MSE_V_ChgRate<0){
textcolor(LIGHTGREEN);
cprintf("%c",DOWN_TRIANGLE);
}
else
{
textcolor(LIGHTRED);
cprintf("%c",UP_TRIANGLE);
}
textcolor(WHITE);


/* Update neurons saturation detector */
if(MSE_T_1==Net->MSE_T){SatDet++;}
else{
if(SatDet){SatDet--;}
}

/* Update validation MSE_lock up detector */
if(MSE_V_1==Net->MSE_V){MSE_V_LockUpDet++;}
else{
if(MSE_V_LockUpDet){MSE_V_LockUpDet--;}
}

/* Count the number of nearly consecutive epochs for which the validation MSE is
reducing (MSE_V_ChgRate<0) in a rate  smaller than Net->MSEChRStopCrit */
if((MSE_V_ChgRate<0)&&(fabs(100*MSE_V_ChgRate)<Net->MSEChRStopCrit)){SmallEnoughMSEChgRateCounter++;}
else{
if(SmallEnoughMSEChgRateCounter){SmallEnoughMSEChgRateCounter--;}
}


/* Terminate training as a consequence of all neurons saturation */
if(SatDet>Net->SmMSEChNEp2C){
QuitPreProcedure(Net,MSEStr);
Quit(SAT_INT,MSEStr);/* training terminated by neurons saturation */
}

/* Convergence is assumed if the number of nearly consecutive epochs for which the
validation MSE is reducing (MSE_V_ChgRate<0) in a rate  smaller than Net->MSEChRStopCrit
is greater than Net->SmMSEChNEp2C OR if validation MSE is locked up */
if((SmallEnoughMSEChgRateCounter>Net->SmMSEChNEp2C)||(MSE_V_LockUpDet>Net->SmMSEChNEp2C)){
QuitPreProcedure(Net,MSEStr);
Quit(SMALL_MSECHG_INT,MSEStr); /* training terminated by small validation MSE change rate */
}


/* Convergence is assumed if validation MSE if small enough */
if(Net->MSE_V<Net->MSEValStopCrit){
QuitPreProcedure(Net,MSEStr);
Quit(SMALL_MSE_INT,MSEStr); /* training terminated by small validation MSE */
}


/* Shuffle Training Set */
RowShuffle(Net->TSet,Net->TSSize);

/* If keyboard hit, call ChangeParameterMenu() menu */
if(kbhit()){
if(ChangeParameterMenu(Net)=='R'){Epoch=Net->NumEpoch;}
ShowParameters(Net);
NetWrite(Net,Net->FRBFSF);
}

/* Update loop variables */
MSE_T_1=Net->MSE_T;
MSE_V_1=Net->MSE_V;
Epoch++;


}/* while(1) */



}


/***************************************************************************
* FUNC: void TSetDeAvg (NEURALNET *Net)
*
* DESC: Subtract the average vector Avg[] from training set TSet.
*       Stores Avg[] in Net. The desired output columns of TSet are
*       left unchanged.
****************************************************************************/
void TSetDeAvg (NEURALNET *Net)
{

register unsigned i,j;



for(j=0;j<Net->NumInputNodes;j++){
Net->Avg[j]=0;

for(i=0;i<Net->TSSize;i++){
Net->Avg[j]+=Net->TSet[i][j];
}

Net->Avg[j]/=Net->TSSize;  /* the average value of the j_th TSet column */

for(i=0;i<Net->TSSize;i++){
Net->TSet[i][j]-=Net->Avg[j]; /* de-averages TSet */
}


}


}


/***************************************************************************
* FUNC: void TSetDeSDev(NEURALNET *Net)
*
* DESC: Normalize each column of training set TSet by its standard deviation
*       SDev. Stores SDev[] in Net. The desired output columns of TSet are
*       left unchanged.
****************************************************************************/
void TSetDeSDev(NEURALNET *Net)
{

register unsigned i,j;



for(j=0;j<Net->NumInputNodes;j++){
Net->SDev[j]=0;

for(i=0;i<Net->TSSize;i++){
Net->SDev[j]+=SQR(Net->TSet[i][j]);
}

Net->SDev[j]/=Net->TSSize;

Net->SDev[j]=sqrt(Net->SDev[j]); /* the standard deviation of the j_th TSet column */

if(Net->SDev[j]){
for(i=0;i<Net->TSSize;i++){
Net->TSet[i][j]/=Net->SDev[j]; /* normalizes TSet by SDev */
}
}


}


}



/*****************************************************************************
* FUNC: void NormalizeDesiredOutput(NEURALNET *Net)
*
* DESC: Determine the desired output set D normalization parameters A_n and B_n,
*       such that Dn=A_n*D+B_n where Dn lies in the range [-1,1]. Then, normalize
*       D to the range [-1,1], i.e., D=Dn.
*****************************************************************************/
void NormalizeDesiredOutput(NEURALNET *Net)
{

register unsigned i;
float v;

float maxv,minv;

maxv=-MAXFLOAT;
minv=MAXFLOAT;

/* determine maxv and minv of desired output set D */
for(i=0;i<Net->TSSize;i++){

v=Net->TSet[i][Net->Rbf.M];
if(v>maxv)maxv=v;
if(v<minv)minv=v;

}

/* determine normalization parameters */
Net->A_n=2.0/(maxv-minv);
Net->B_n=(maxv+minv)/(minv-maxv);

/* normalize desired output set */
for(i=0;i<Net->TSSize;i++){

Net->TSet[i][Net->Rbf.M]=Net->A_n*Net->TSet[i][Net->Rbf.M]+Net->B_n;

}



}


/***************************************************************************
* FUNC: void VSetDeAvg (NEURALNET *Net)
*
* DESC: Subtract the average vector Avg[] from training validation VSet.
*       The desired output columns of VSet are left unchanged.
****************************************************************************/
void VSetDeAvg (NEURALNET *Net)
{

register unsigned i,j;


for(j=0;j<Net->NumInputNodes;j++){

for(i=0;i<Net->VSSize;i++){
Net->VSet[i][j]-=Net->Avg[j]; /* de-averages VSet */
}


}


}


/***************************************************************************
* FUNC: void VSetDeSDev(NEURALNET *Net)
*
* DESC: Normalize each column of validation set VSet by the corresponding
*       column standard deviation SDev in TSet. The desired output columns
*       of VSet are left unchanged.
****************************************************************************/
void VSetDeSDev(NEURALNET *Net)
{

register unsigned i,j;


for(j=0;j<Net->NumInputNodes;j++){


if(Net->SDev[j]){
for(i=0;i<Net->VSSize;i++){
Net->VSet[i][j]/=Net->SDev[j]; /* normalize VSet by TSet SDev */
}
}


}


}


/*****************************************************************************
* FUNC: void NormalizeVSetDesiredOutput(NEURALNET *Net)
*
* DESC: Normalize VSet desired output set to the range [-1,1].
*****************************************************************************/
void NormalizeDesiredVSetOutput(NEURALNET *Net)
{

register unsigned i;

/* normalize VSet desired output set */
for(i=0;i<Net->VSSize;i++){

Net->VSet[i][Net->Rbf.M]=Net->A_n*Net->VSet[i][Net->Rbf.M]+Net->B_n;

}



}


/*****************************************************************************
* FUNC: void QuitPreProcedure(NEURALNET *Net, char *MSE2String)
*
* DESC: Perform pre-exit procedure before the call of Quit() due to training/validation
*       termination.
*****************************************************************************/
void QuitPreProcedure(NEURALNET *Net, char *MSE2String)

{

register unsigned i;
unsigned nrows,ncols_;

/* close MSE output file */
fclose(Net->MSEOFp);

/* save specification after training */
NetWrite(Net,Net->FRBFSF);

/* save Net->MSE_V into MSE2String */
sprintf(MSE2String," %-.3E",Net->MSE_V);

if(!Net->ValidationMode){ /* -> Net->VSet==Net->TSet */

/* Re-Read Training Set (to circumvent RowShuffle() action) */
Net->TSet=FloatMatReadCheck(Net->TSet,Net->TDIF,Net->buffer,&nrows,&ncols_,0);

/*  Normalize TSet(=VSet) desired output set to range [-1,1] */
NormalizeDesiredVSetOutput(Net);

/* De-averages Training Set */
VSetDeAvg(Net);

/* Normalizes each TSet input vector component by the corresponding
component of the TSet standard deviation vector. */
VSetDeSDev(Net);

}

/* Open RBF output neurons output file */
if((Net->RBFO2VOFp=fopen(Net->RBFO2VOF,"wt"))==NULL)Quit(WRITE_OPEN_ERR,Net->RBFO2VOF);
if (setvbuf(Net->RBFO2VOFp,Net->buffer,_IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Net->RBFO2VOF);


/* Validation Set presentation */
for(i=0;i<Net->VSSize;i++){

RBFOutVal(Net,Net->VSet[i]);

/* Save Output Neuron values */
fprintf(Net->RBFO2VOFp,"%8.6G\n",(Net->Rbf.y-Net->B_n)/Net->A_n);
}

fclose(Net->RBFO2VOFp);

#if(0)
putch(BEL);
#endif

}


/****************************************************************************
* FUNC: inline float FloatRandGen(float Range)
*
* DESC: Generate float random numbers in the range [-Range, +Range] with
*       uniform probabilistic distribution.
*
*****************************************************************************/
inline float FloatRandGen(float Range)

{

return  (((float)rand()/(RAND_MAX/2))-1)*Range;

}


/*****************************************************************************
* FUNC: inline float GetMinAbsValue(float **VectorSet,unsigned NumVect,unsigned VectDim)
*
* DESC: Determine the minimum absolute value component for the
*       the vectors stored in VectorSet.
******************************************************************************/
inline float GetMinAbsValue(float **VectorSet,unsigned NumVect,unsigned VectDim)
{
register unsigned i,j;
float Min=MAXFLOAT;

/* Get Min */
for(i=0;i<NumVect;i++)
for(j=0;j<VectDim;j++)
if(fabs(VectorSet[i][j])<Min)Min=fabs(VectorSet[i][j]);

return Min;

}



/*****************************************************************************
* FUNC: inline float GetMaxAbsValue(float **VectorSet,unsigned NumVect,unsigned VectDim)
*
* DESC: Determine the maximum absolute value component for the
*       the vectors stored in VectorSet.
******************************************************************************/
inline float GetMaxAbsValue(float **VectorSet,unsigned NumVect,unsigned VectDim)
{
register unsigned i,j;
float Max=MINABSFLOAT;

/* Get Max */
for(i=0;i<NumVect;i++)
for(j=0;j<VectDim;j++)
if(fabs(VectorSet[i][j])>Max)Max=fabs(VectorSet[i][j]);

return Max;

}


/***************************************************************************
* FUNC: inline float DiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim)
*
* DESC: Return the squared Euclidean norm of Vector0-Vector1.
****************************************************************************/
inline float DiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim)
{

register unsigned  i;

float SqEuclNorm=0;

for(i=0;i<VectorDim;i++) SqEuclNorm+=SQR(Vector0[i]-Vector1[i]);

return SqEuclNorm;


}


/***************************************************************************
* FUNC: inline float SquaredEuclideanNorm(float *Vector,unsigned VectorDim)
*
* DESC: Determine the squared Euclidean norm of vector Vector.
****************************************************************************/
inline float SquaredEuclideanNorm(float *Vector,unsigned VectorDim)
{

register unsigned i;

float SqEuclNorm=0;

for(i=0;i<VectorDim;i++) SqEuclNorm+=SQR(Vector[i]);

return SqEuclNorm;


}

/***************************************************************************
* FUNC: void KMeans(NEURALNET *Net)
*
* DESC: KMeans algorithm (Kohonen Neural network with unit neighborhood function)
****************************************************************************/
void KMeans(NEURALNET *Net)
{
#define KMEANS_MAXITER 65536

register unsigned Epoch,j,k;

float MinSqEuclNorm, AvgSqEuclNorm,PrevAvgSqEuclNorm,SqEuclNormChg,SmoothSqEuclNormChg;
float Tmp;
unsigned Index,k_tilde=0;
float StopCriterion;
char op;


if(!Net->Kmeans_On){/* K-means is deactivated  */
Net->K_Means_NumEpoch=0;
return;
}

/* Init with zero all K Vector Center Average Vectors taux[] */
for(k=0;k<Net->Rbf.K;k++) {
for(j=0;j<Net->Rbf.M;j++) {
Net->Rbf.Hn[k].taux[j]=0.0;
}
}

Epoch=0;/* epoch counter */


AvgSqEuclNorm=SmoothSqEuclNormChg=SqEuclNormChg=0;
StopCriterion=Net->Rbf.Kn*(Net->Rbf.KMStopCrit/100);

/* Epoch loop */
do{

PrevAvgSqEuclNorm=AvgSqEuclNorm;
AvgSqEuclNorm=0;

/* Initial procedures */
gotoxy(XPos[0],YPos[0]);cprintf("%u",Epoch); /* show in Status screen */


/* TSet loop */
for(j=0;j<Net->Rbf.N;j++){

/* Randomly picks a vector from TSet */
Index=(unsigned)rand()%(Net->Rbf.N);

/* Determine k_tilde for this vector */
MinSqEuclNorm=MAXFLOAT;
for(k=0;k<Net->Rbf.K;k++){
Tmp=DiffSquaredEuclideanNorm(Net->TSet[Index],Net->Rbf.Hn[k].t,Net->Rbf.M);
if(Tmp<MinSqEuclNorm){MinSqEuclNorm=Tmp; k_tilde=k;}
} /* for k */

/* Update Vector Center k_tilde */
for(k=0;k<Net->Rbf.M;k++)

Net->Rbf.Hn[k_tilde].t[k]+=Net->Rbf.Kn*(Net->TSet[Index][k]-Net->Rbf.Hn[k_tilde].t[k]);

}/* End of TSet loop - for j */

/* Init procedures for the stopping criterion check: */

/* Determine all K Vector Center Average Vectors taux[] for the present Epoch */
for(k=0;k<Net->Rbf.K;k++) {
for(j=0;j<Net->Rbf.M;j++) {
Net->Rbf.Hn[k].taux[j]=Net->Rbf.Hn[k].taux[j]*((float)Epoch/(Epoch+1))+(Net->Rbf.Hn[k].t[j])/(Epoch+1);
}
}

/* Determine the average squared Euclidean norm of all K Vector Center Average Vectors */
for(k=0;k<Net->Rbf.K;k++) {
AvgSqEuclNorm+=SquaredEuclideanNorm(Net->Rbf.Hn[k].taux,Net->Rbf.M);
}
AvgSqEuclNorm/=Net->Rbf.K;


/* Determine the change on AvgSqEuclNorm from previous to present Epoch */
if(PrevAvgSqEuclNorm!=0){
SqEuclNormChg=fabs((AvgSqEuclNorm/PrevAvgSqEuclNorm)-1);
}

/* Store SqEuclNormChg value for the present Epoch in AuxK[] auxiliary vector,
such that AuxK[] is an array of the present epoch SqEuclNormChg value plus
the K-1 ones */
Net->Rbf.AuxK[Epoch%Net->Rbf.K]=SqEuclNormChg;

/* Average the K SqEuclNormChg values stored in AuxK[] in a continuous basis
   and save the resulting scalar into SmoothSqEuclNormChg */
if(Epoch>=((Net->Rbf.K)-1)){
for(k=0;k<Net->Rbf.K;k++)SmoothSqEuclNormChg=SmoothSqEuclNormChg*((float)k/(k+1))+(Net->Rbf.AuxK[k])/(k+1);
}

Epoch++;/* increments epoch counter */

if(Epoch>=KMEANS_MAXITER)Quit(MAXIT_ERR,"KMeans()");


/* If keyboard hit, call ChangeParameterMenu() menu */
if(kbhit()){
op=ChangeParameterMenu(Net);
if(op=='B')break; /* break do-while loop and go to Stochastic Gradient */
ShowParameters(Net);
}


}while((Epoch<Net->Rbf.K+1)||(SmoothSqEuclNormChg>StopCriterion));/* stopping (convergence) criterion - Note: the inverse logic of the condition */
/* End of Epoch loop */


Net->Kmeans_On=0;/* K-means is terminated */

/* save number of epochs  */
Net->K_Means_NumEpoch=Epoch;

/* save the Vector Centers determined by K-Means  */
for(k=0;k<Net->Rbf.K;k++) {
for(j=0;j<Net->Rbf.M;j++) {
Net->Rbf.Hn[k].taux[j]=Net->Rbf.Hn[k].t[j];
}
}


}



/***************************************************************************
* FUNC: float MaxSqEuclNormBetweenCenters(NEURALNET *Net)
*
* DESC: Returns the square of the max distance between centers.
****************************************************************************/
float MaxSqEuclNormBetweenCenters(NEURALNET *Net)
{

float MaxDist2,tmp;
register unsigned i,j;

MaxDist2=10*MINABSFLOAT;;

	for(i=0;i<Net->Rbf.K;i++){
	for(j=i+1;j<Net->Rbf.K;j++){
		tmp=DiffSquaredEuclideanNorm(Net->Rbf.Hn[i].t,Net->Rbf.Hn[j].t,Net->Rbf.M);
		if(tmp>MaxDist2){MaxDist2=tmp;}
	}
	}
return MaxDist2;

}

/***************************************************************************
* FUNC: inline void Phy_O(NEURALNET *Net,float *InputVector)
*
* DESC: Determine the output value of each hidden neuron.
****************************************************************************/
inline void Phy_O(NEURALNET *Net,float *InputVector)
{

register unsigned k;

for(k=0;k < Net->Rbf.K;k++){

if(Net->Rbf.Hn[k].ss<=0.0)Net->Rbf.Hn[k].ss=10*MINABSFLOAT;

Net->Rbf.Hn[k].NDN= AULIM(DiffSquaredEuclideanNorm(InputVector,Net->Rbf.Hn[k].t,Net->Rbf.M)/Net->Rbf.Hn[k].ss);

Net->Rbf.Hn[k].phy=exp(-Net->Rbf.Hn[k].NDN);

}

}

/***************************************************************************
* FUNC: inline float RBFOutVal(NEURALNET *Net,float *InputVector)
*
* DESC: Determine the RBF output value y for the n_th TSet vector
        and return the resulting squared error.
****************************************************************************/
inline float RBFOutVal(NEURALNET *Net,float *InputVector)
{

register unsigned k;

/* Determine the Hidden Neurons (RBF Centers) output (Net->Hn[k].phy) to the
 InputVector training vector */

Phy_O(Net,InputVector);

/* Determine the RBF Neural Net output y */
Net->Rbf.y=0;

for(k=0;k<Net->Rbf.K;k++){
Net->Rbf.y+=Net->Rbf.W[k]*Net->Rbf.Hn[k].phy;
}

Net->Rbf.y+=Net->Rbf.Wb*BIAS;

 /* Determine the error e=d-y */
Net->Rbf.e=InputVector[Net->Rbf.M]-(Net->Rbf.y);

/* return the instant squared error */
return SQR(Net->Rbf.e);

}


/***************************************************************************
* FUNC: inline void StochGrad(NEURALNET *Net,float *InputVector)
*
* DESC: RBF free-parameters adjustment via Stochastic Gradient Algorithm
* NOTE: See Adaptive Filters - Simon Haykin - Chapter 20 - Table 20.1
****************************************************************************/
inline void StochGrad(NEURALNET *Net,float *InputVector)
{

register unsigned j,k;

/* ---- Vector Centers update procedure: ---- */

for(k=0;k<Net->Rbf.K;k++){

/* Determine the difference between InputVector and the k_th Vector Center,
   then normalize the resulting vector w.r.t the k_th center variance */
for(j=0;j<Net->Rbf.M;j++){
Net->Rbf.AuxM[j]= AULIM((InputVector[j]-Net->Rbf.Hn[k].t[j])/Net->Rbf.Hn[k].ss);
}

/* k_th Vector Center update */
for(j=0;j<Net->Rbf.M;j++){
Net->Rbf.Hn[k].t[j]+=2*Net->Rbf.Tn*Net->Rbf.AuxM[j]*Net->Rbf.Hn[k].phy*Net->Rbf.W[k]*Net->Rbf.e;
}

}/* for k */


/* ---- Centers Variance update procedure: ---- */


for(k=0;k<Net->Rbf.K;k++){
Net->Rbf.Hn[k].ss+=Net->Rbf.Sn*AULIM(Net->Rbf.Hn[k].NDN/Net->Rbf.Hn[k].ss)*Net->Rbf.Hn[k].phy*Net->Rbf.W[k]*Net->Rbf.e;
}


/* ---- Output Neuron Synapses update procedure: ---- */

for(k=0;k<Net->Rbf.K;k++){
Net->Rbf.W[k]+=Net->Rbf.Wn*Net->Rbf.e*Net->Rbf.Hn[k].phy;
}

Net->Rbf.Wb+=Net->Rbf.Wn*Net->Rbf.e*BIAS;

}




/***************************************************************************
* FUNC: inline void InitRBFVectorCenters(NEURALNET *Net)
*
* DESC: Randomly init RBF Vector Centers.
****************************************************************************/
inline void InitRBFVectorCenters(NEURALNET *Net)

{

register unsigned j,k;
float Min,Max;

if(Net->Rbf.Kn==0.0){ /* init procedure if K-means learning rate == 0.0 */

/* Randomly init the components of all K Vector Centers - uniform
distribution between [-Max,Max] where Max is the
training set maximum absolute value component: */

Max=GetMaxAbsValue(Net->TSet,Net->TSSize,Net->Rbf.M);
//printf("Max=%f\n",Max);

/* save the Vector Centers randomly initialized by K-Means  */
for(k=0;k<Net->Rbf.K;k++) {
for(j=0;j<Net->Rbf.M;j++) {
Net->Rbf.Hn[k].taux[j]=Net->Rbf.Hn[k].t[j]=FloatRandGen(Max);
}
}

#ifdef DEBUG_FILE
for(k=0;k < Net->Rbf.K;k++){
for(j=0;j < Net->Rbf.M;j++){
fprintf(Dbg,"%f\t",Net->Rbf.Hn[k].t[j]);
}
fprintf(Dbg,"\n");
}
#endif

Net->Kmeans_On=0;/* K-means is deactivated  */

}
else{/* init procedure if K-means learning rate != 0.0 */

/* Randomly init the components of all K Vector Centers - uniform
distribution between [-0.1*Min,0.1*Min] where Min is the
training set minimum absolute value component: */

/* Get the TSet minimum absolute value component */
Min=GetMinAbsValue(Net->TSet,Net->TSSize,Net->Rbf.M);
//printf("Min=%f\n",Min);

Min=0.1*Min;
if(Min==0.0)Min=10*MINABSFLOAT;

for(k=0;k<Net->Rbf.K;k++) {
for(j=0;j<Net->Rbf.M;j++) {
Net->Rbf.Hn[k].t[j]=FloatRandGen(Min);
}
}

#ifdef DEBUG_FILE
for(k=0;k < Net->Rbf.K;k++){
for(j=0;j < Net->Rbf.M;j++){
fprintf(Dbg,"%f\t",Net->Rbf.Hn[k].t[j]);
}
fprintf(Dbg,"\n");
}
#endif

Net->Kmeans_On=1;/* K-means is activated  */

}


}
